
/*
 *                            COPYRIGHT
 *
 *  XAnalyser, frequence spectrum analyser for X Windows
 *  Copyright (C) 1998 Arvin Schnell
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Contact addresses:
 *  aschnell@physik.uni-bremen.de or arvinschnell@yahoo.com
 *  Arvin Schnell, Am Heidberg 8, 28865 Lilienthal, Germany
 *
 */


#include <stdlib.h>
#include <iostream.h>
#include <iomanip.h>
#include <builtin.h>

#include <X11/Xlib.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <Xm/Protocols.h>

#include "Trans.hh"
#include "main.hh"
#include "display.hh"
#include "control.hh"
#include "calc.hh"
#include "misc.hh"


Widget display_w;

Display *display;
int screen;
Colormap colormap;

Widget frame_w[2];
Widget analyser_w[2];

Window window[2];

unsigned int width, height;

GC gc;

int font_width, font_height;

Trans2<float,short> K;

int visibility;




void
display_draw( XRectangle rect, int c, int complete = 1 )
{
  if( !visibility )
    return;
  
  XSetClipRectangles( display, gc, 0, 0, &rect, 1, Unsorted );
  
  int first = rect.x;
  int last = rect.width + rect.x;
  
  if( complete ) {
    
    XSetForeground( display, gc, xanalyser.backgroundcolor );
    XFillRectangle( display, window[c], gc, 0, 0, width, height );
    
    XSetForeground( display, gc, xanalyser.datacolor );
    for( int i = first; i < last; i++ )
      XDrawLine( display, window[c], gc, i, value_y[c][i], i, height-1 );
    
  } else {
    
    XSetForeground( display, gc, xanalyser.backgroundcolor );
    for( int i = first; i < last; i++ )
      if( value_y[c][i] > value_y_old[c][i] )
	XDrawLine( display, window[c], gc, i, value_y_old[c][i],
		   i, value_y[c][i]-1 );
    
    XSetForeground( display, gc, xanalyser.datacolor );
    for( int i = first; i < last; i++ )
      if( value_y[c][i] < value_y_old[c][i] )
	XDrawLine( display, window[c], gc, i, value_y_old[c][i],
		   i, value_y[c][i] );
    
  }
  
  
  for( int f = 0; f < 25000; f += 2500 ) {
    
    short x;
    K.X.AtoB( (float)(f), x );
    
    if( f % 1000 == 0 )
      XSetForeground( display, gc, xanalyser.majorgridcolor );
    else
      XSetForeground( display, gc, xanalyser.minorgridcolor );
    
    if( f > 0 )
      XDrawLine( display, window[c], gc, x, 0, x, height-1 );
    
    if( complete || frame_count % (int)(frames_per_second) == 0 ) {
      const int size = 10;
      char buffer[size];
      sprintf( buffer, "%3.1f", (double)(f)/1000.0 );
      XDrawString( display, window[c], gc, x+3, font_height,
		   buffer, strlen( buffer ) );
    }
  }
  
  
  for( int db = (int)( 20.0 * ceil( upper_db/20.0 ) ); db > lower_db; db += -20 ) {
    
    short y;
    K.Y.AtoB( (float)(db), y );
    
    if( db == 0 )
      XSetForeground( display, gc, xanalyser.majorgridcolor );
    else
      XSetForeground( display, gc, xanalyser.minorgridcolor );
    
    XDrawLine( display, window[c], gc, 0, y, width-1, y );
    
    if( complete || frame_count % (int)(frames_per_second) == 0 ) {
      const int size = 10;
      char buffer[size];
      sprintf( buffer, "%+d", db );
      XDrawString( display, window[c], gc, width-2 - font_width*strlen( buffer ),
		   y-2, buffer, strlen( buffer ) );
    }
  }
  
  
  if( searchon )
    for( int m = 0; m < 2; m++ ) {
      int x = marker_x[m];
      XSetForeground( display, gc, xanalyser.markercolor );
      XDrawLine( display, window[c], gc, x, 0, x, height-1 );
    }
  
}



// update displays

void
display_update()
{
  if( !display_w )
    return;
  
  for( int c = 0; c < num_channels; c++ ) {
    XRectangle rect = { 0, 0, width, height };
    display_draw( rect, c, 0 );
  }
  
  XFlush( display );
}



// draw or remove markers

void
display_marker()
{
  if( !display_w )
    return;
  
  for( int c = 0; c < num_channels; c++ )
    for( int m = 0; m < 2; m++ ) {
      XRectangle rect = { marker_x[m], 0, 1, height };
      display_draw( rect, c );
    }
}



void
display_realize()
{
  // we don't want to do this more than one time
  
  static done = 0;
  if( done ) return;
  done = 1;
  
  
  // get display, screen, window, colormap
  
  display = XtDisplay( analyser_w[0] );
  screen = DefaultScreen( display );
  colormap = DefaultColormap( display, screen );
  
  for( int c = 0; c < 2; c++ )
    window[c] = XtWindow( analyser_w[c] );
  
  
  // create GCs
  
  XtGCMask gc_mask = 0;
  XGCValues gc_values;
  
  gc = XCreateGC( display, window[0], gc_mask, &gc_values );
  
  
  // get width and height (is equal for both windows)
  
  myXGetDrawableSize( display, window[0], &width, &height );
  
  
  // font
  
  XSetFont( display, gc, xanalyser.gridfont->fid );
  font_width = xanalyser.gridfont->max_bounds.rbearing
    - xanalyser.gridfont->min_bounds.lbearing;
  font_height = xanalyser.gridfont->max_bounds.ascent
    + xanalyser.gridfont->max_bounds.descent;
  
  
  // trans - other part in calc_init()
  
  K.X.setB( 0, width-1 );
  K.Y.setB( height-1, 0 );
  
  
  // clear stuff
  
  for( int c = 0; c < 2; c++ )
    for( int i = 0; i < num_fft; i++ ) {
      K.Y.AtoB( lower_db, value_y[c][i] );
      value_y_old[c][i] = value_y[c][i];
    }
  
  
  // misc
  
  marker_x[0] = 0;
  marker_x[1] = width - 1;
  
}





void
set_marker( int x )
{
   // which marker is closer?
  
  int m = 0;
  if( abs( marker_x[0] - x ) > abs( marker_x[1] - x ) )
    m = 1;
  
  int old_x = marker_x[m];
  marker_x[m] = x;
  
  
  // check positions
  
  if( marker_x[0] > marker_x[1] ) {
    int tmp = marker_x[0];
    marker_x[0] = marker_x[1];
    marker_x[1] = tmp;
  }
  
  if( marker_x[0] < 0 )
    marker_x[0] = 0;
  
  if( marker_x[1] > num_fft-1 )
    marker_x[1] = num_fft-1;
  
  
  // remove old marker
  
  for( int c = 0; c < num_channels; c++ ) {
    XRectangle rect = { old_x, 0, 1, height };
    display_draw( rect, c );
  }
  
  
  // draw new marker and update label
  
  for( int c = 0; c < num_channels; c++ ) {
    XRectangle rect = { marker_x[m], 0, 1, height };
    display_draw( rect, c );
  }
  
  
  calc_search_peak();
  update_marker_label();
  update_peak_label();
}



void
display_callback( Widget, XtPointer client_data, XtPointer call_data )
{
  XmDrawingAreaCallbackStruct *c = (XmDrawingAreaCallbackStruct *) call_data;
  
  switch( c->reason ) {
    
  case XmCR_EXPOSE: {
    
    display_realize();  // misuse as realise callback
    
    XExposeEvent *e = (XExposeEvent *) c->event;
    XRectangle rect = { e->x, e->y, e->width, e->height };
    
    int c = (int) client_data;
    display_draw( rect, c );
    
  } break;
  
  case XmCR_INPUT: {
    
    switch( c->event->type ) {
      
    case ButtonPress:
    case MotionNotify: {
      XButtonPressedEvent *e = (XButtonPressedEvent *) c->event;
      XPoint point = { e->x, e->y };
      
      if( searchon )
	set_marker( point.x );
      
    } break;
    
    }
    
  } break;
  
  }
}



void
display_notify( Widget, XtPointer, XEvent * event, Boolean * )
{
  switch( event->type ) {
    
  case UnmapNotify:
    // cout << "UnmapNotify\n";
    visibility = 0;
    break;
    
  case VisibilityNotify:
    // cout << "VisibilityNotify\n";
    visibility = ((XVisibilityEvent*)(event))->state != VisibilityFullyObscured;
    break;
    
  }
}



void
display_create()
{
  Arg args[10];
  int n;
  
  
  // form
  
  n = 0;
  XtSetArg( args[n], XtNiconPixmap, icon_pixmap ); n++;
  XtSetArg( args[n], XtNiconMask, icon_mask ); n++;
  display_w = XmCreateFormDialog( toplevel, "Display", args, n );
  XmAddWMProtocolCallback( XtParent( display_w ), wm_delete_window, quit_cb, NULL );
  XtAddEventHandler( XtParent( display_w ), StructureNotifyMask, False,
		     display_notify, NULL );
  XtAddEventHandler( XtParent( display_w ), VisibilityChangeMask, False,
		     display_notify, NULL );
  
  
  // frame 0
  
  n = 0;
  XtSetArg( args[n], XmNtopAttachment, XmATTACH_FORM ); n++;
  XtSetArg( args[n], XmNleftAttachment, XmATTACH_FORM ); n++;
  frame_w[0] = XmCreateFrame( display_w, "Frame0", args, n );
  XtManageChild( frame_w[0] );
  
  
  // drawingarea 0
  
  n = 0;
  XtSetArg( args[n], XmNwidth, num_fft ); n++;
  analyser_w[0] = XmCreateDrawingArea( frame_w[0], "Analyser0", args, n );
  XtManageChild( analyser_w[0] );
  XtAddCallback( analyser_w[0], XmNexposeCallback, display_callback,
		 (XtPointer) 0 );
  XtAddCallback( analyser_w[0], XmNinputCallback, display_callback,
		 (XtPointer) 0 );
  
  
  // frame 1
  
  n = 0;
  XtSetArg( args[n], XmNtopAttachment, XmATTACH_WIDGET ); n++;
  XtSetArg( args[n], XmNtopWidget, frame_w[0] ); n++;
  XtSetArg( args[n], XmNleftAttachment, XmATTACH_FORM ); n++;
  frame_w[1] = XmCreateFrame( display_w, "Frame1", args, n );
  XtManageChild( frame_w[1] );
  
  
  // drawingarea 1
  
  n = 0;
  XtSetArg( args[n], XmNwidth, num_fft ); n++;
  analyser_w[1] = XmCreateDrawingArea( frame_w[1], "Analyser1", args, n );
  if( num_channels == 2 ) XtManageChild( analyser_w[1] );
  XtAddCallback( analyser_w[1], XmNexposeCallback, display_callback,
		 (XtPointer) 1 );
  XtAddCallback( analyser_w[1], XmNinputCallback, display_callback,
		 (XtPointer) 1 );
}



void
display_manage()
{
  if( !display_w )
    display_create();
  
  if( !XtIsManaged( display_w ) )
    XtManageChild( display_w );
  else
    XMapRaised( XtDisplay( display_w ), XtWindow( XtParent( display_w ) ) );
}



void
display_remanage()
{
  if( !display_w )
    return;
  
  if( num_channels == 1 ) {
    XtUnmanageChild( frame_w[1] );
    XtUnmanageChild( analyser_w[1] );
  } else {
    XtManageChild( frame_w[1] );
    XtManageChild( analyser_w[1] );
  }
}

